@@ -31,42 +31,41 @@
 #include <X11/Xlib.h>
 #include <X11/Xatom.h>
 #include <X11/extensions/XInput.h>
-#ifdef HAVE_XRECORD
+#ifdef HAVE_X11_EXTENSIONS_RECORD_H
 #include <X11/Xproto.h>
 #include <X11/extensions/record.h>
-#endif /* HAVE_XRECORD */
+#endif /* HAVE_X11_EXTENSIONS_RECORD_H */
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <signal.h>
-#include <sys/ipc.h>
-#include <sys/shm.h>
 #include <sys/time.h>
 #include <sys/stat.h>
 
 #include "synaptics.h"
 #include "synaptics-properties.h"
 
-enum TouchpadState {
+typedef enum {
     TouchpadOn = 0,
     TouchpadOff = 1,
     TappingOff = 2
-};
+} TouchpadState;
 
 
-static SynapticsSHM *synshm;
-static int pad_disabled;
-static int disable_taps_only;
+static Bool pad_disabled /* internal flag, this does not correspond to device state */;
 static int ignore_modifier_combos;
 static int ignore_modifier_keys;
 static int background;
 static const char *pid_file;
-static int use_shm;
 static Display *display;
 static XDevice *dev;
 static Atom touchpad_off_prop;
+static TouchpadState previous_state;
+static TouchpadState disable_state = TouchpadOff;
+static int verbose;
 
 #define KEYMAP_SIZE 32
 static unsigned char keyboard_mask[KEYMAP_SIZE];
@@ -85,42 +84,58 @@ usage(void)
     fprintf(stderr, "  -k Ignore modifier keys when monitoring keyboard activity.\n");
     fprintf(stderr, "  -K Like -k but also ignore Modifier+Key combos.\n");
     fprintf(stderr, "  -R Use the XRecord extension.\n");
+    fprintf(stderr, "  -v Print diagnostic messages.\n");
     exit(1);
 }
 
+static void
+store_current_touchpad_state(void)
+{
+    Atom real_type;
+    int real_format;
+    unsigned long nitems, bytes_after;
+    unsigned char *data;
+
+    if ((XGetDeviceProperty (display, dev, touchpad_off_prop, 0, 1, False,
+                             XA_INTEGER, &real_type, &real_format, &nitems,
+                             &bytes_after, &data) == Success) && (real_type != None)) {
+        previous_state = data[0];
+    }
+}
+
 /**
  * Toggle touchpad enabled/disabled state, decided by value.
  */
 static void
-toggle_touchpad(enum TouchpadState value)
+toggle_touchpad(Bool enable)
 {
-    if (pad_disabled && !value)
-    {
-        if (!background)
+    unsigned char data;
+    if (pad_disabled && enable) {
+        data = previous_state;
+        pad_disabled = False;
+        if (verbose)
             printf("Enable\n");
-    } else if (!pad_disabled && value)
-    {
-        if (!background)
+    } else if (!pad_disabled && !enable &&
+               previous_state != disable_state &&
+               previous_state != TouchpadOff) {
+        store_current_touchpad_state();
+        pad_disabled = True;
+        data = disable_state;
+        if (verbose)
             printf("Disable\n");
     } else
         return;
 
-    pad_disabled = value;
-    if (use_shm)
-        synshm->touchpad_off = value;
-    else {
-        unsigned char data = value;
-        /* This potentially overwrites a different client's setting, but ...*/
-	XChangeDeviceProperty(display, dev, touchpad_off_prop, XA_INTEGER, 8,
-				PropModeReplace, &data, 1);
-	XFlush(display);
-    }
+    /* This potentially overwrites a different client's setting, but ...*/
+    XChangeDeviceProperty(display, dev, touchpad_off_prop, XA_INTEGER, 8,
+            PropModeReplace, &data, 1);
+    XFlush(display);
 }
 
 static void
 signal_handler(int signum)
 {
-    toggle_touchpad(TouchpadOn);
+    toggle_touchpad(True);
 
     if (pid_file)
 	unlink(pid_file);
@@ -191,24 +206,6 @@ keyboard_activity(Display *display)
     return ret;
 }
 
-/**
- * Return non-zero if any physical touchpad button is currently pressed.
- */
-static int
-touchpad_buttons_active(void)
-{
-    int i;
-
-    if (synshm->left || synshm->right || synshm->up || synshm->down)
-	return 1;
-    for (i = 0; i < 8; i++)
-	if (synshm->multi[i])
-	    return 1;
-    if (synshm->guest_left || synshm->guest_mid || synshm->guest_right)
-        return 1;
-    return 0;
-}
-
 static double
 get_time(void)
 {
@@ -223,20 +220,22 @@ main_loop(Display *display, double idle_time, int poll_delay)
     double last_activity = 0.0;
     double current_time;
 
-    pad_disabled = 0;
     keyboard_activity(display);
 
     for (;;) {
 	current_time = get_time();
 	if (keyboard_activity(display))
 	    last_activity = current_time;
-	if (use_shm && touchpad_buttons_active())
-	    last_activity = 0.0;
+
+	/* If system times goes backwards, touchpad can get locked. Make
+	 * sure our last activity wasn't in the future and reset if it was. */
+	if (last_activity > current_time)
+	    last_activity = current_time - idle_time - 1;
 
 	if (current_time > last_activity + idle_time) {	/* Enable touchpad */
-	    toggle_touchpad(TouchpadOn);
+	    toggle_touchpad(True);
 	} else {			    /* Disable touchpad */
-	    toggle_touchpad(disable_taps_only ? TappingOff : TouchpadOff);
+	    toggle_touchpad(False);
 	}
 
 	usleep(poll_delay);
@@ -272,7 +271,7 @@ setup_keyboard_mask(Display *display, int ignore_modifier_keys)
 }
 
 /* ---- the following code is for using the xrecord extension ----- */
-#ifdef HAVE_XRECORD
+#ifdef HAVE_X11_EXTENSIONS_RECORD_H
 
 #define MAX_MODIFIERS 16
 
@@ -299,7 +298,7 @@ Bool check_xrecord(Display *display) {
 			    &first_error);
 
     status = XRecordQueryVersion(display, version, version+1);
-    if (!background && status) {
+    if (verbose && status) {
 	printf("X RECORD extension version %d.%d\n", version[0], version[1]);
     }
     return found;
@@ -383,8 +382,6 @@ void record_main_loop(Display* display, double idle_time) {
     XRecordRange *range;
     int i;
 
-    pad_disabled = 0;
-
     dpy_data = XOpenDisplay(NULL); /* we need an additional data connection. */
     range  = XRecordAllocRange();
 
@@ -439,19 +436,19 @@ void record_main_loop(Display* display, double idle_time) {
 	    timeout.tv_sec  = (int)idle_time;
 	    timeout.tv_usec = (idle_time-(double)timeout.tv_sec) * 1.e6;
 
-	    toggle_touchpad(disable_taps_only ? TappingOff : TouchpadOff);
+	    toggle_touchpad(False);
 	}
 
 	if (ret == 0 && pad_disabled) { /* timeout => enable event */
-	    toggle_touchpad(TouchpadOn);
-	    if (!background) printf("enable touchpad\n");
+	    toggle_touchpad(True);
+	    if (verbose) printf("enable touchpad\n");
 	}
 
     } /* end while(1) */
 
     XFreeModifiermap(cbres.modifiers);
 }
-#endif /* HAVE_XRECORD */
+#endif /* HAVE_X11_EXTENSIONS_RECORD_H */
 
 static XDevice *
 dp_get_device(Display *dpy)
@@ -518,28 +515,6 @@ unwind:
     return dev;
 }
 
-static int
-shm_init()
-{
-    int shmid;
-
-    /* Connect to the shared memory area */
-    if ((shmid = shmget(SHM_SYNAPTICS, sizeof(SynapticsSHM), 0)) == -1) {
-	if ((shmid = shmget(SHM_SYNAPTICS, 0, 0)) == -1) {
-	    fprintf(stderr, "Can't access shared memory area. SHMConfig disabled?\n");
-	    return 0;
-	} else {
-	    fprintf(stderr, "Incorrect size of shared memory area. Incompatible driver version?\n");
-	    return 0;
-	}
-    }
-    if ((synshm = (SynapticsSHM*) shmat(shmid, NULL, 0)) == NULL) {
-	perror("shmat");
-	return 0;
-    }
-    return 1;
-}
-
 int
 main(int argc, char *argv[])
 {
@@ -549,7 +524,7 @@ main(int argc, char *argv[])
     int use_xrecord = 0;
 
     /* Parse command line parameters */
-    while ((c = getopt(argc, argv, "i:m:dtp:kKR?")) != EOF) {
+    while ((c = getopt(argc, argv, "i:m:dtp:kKR?v")) != EOF) {
 	switch(c) {
 	case 'i':
 	    idle_time = atof(optarg);
@@ -561,7 +536,7 @@ main(int argc, char *argv[])
 	    background = 1;
 	    break;
 	case 't':
-	    disable_taps_only = 1;
+	    disable_state = TappingOff;
 	    break;
 	case 'p':
 	    pid_file = optarg;
@@ -573,12 +548,12 @@ main(int argc, char *argv[])
 	    ignore_modifier_combos = 1;
 	    ignore_modifier_keys = 1;
 	    break;
-        case 's':
-            use_shm = 1;
-            break;
 	case 'R':
 	    use_xrecord = 1;
 	    break;
+	case 'v':
+	    verbose = 1;
+	    break;
 	default:
 	    usage();
 	    break;
@@ -594,9 +569,7 @@ main(int argc, char *argv[])
 	exit(2);
     }
 
-    if (use_shm && !shm_init())
-	exit(2);
-    else if (!use_shm && !(dev = dp_get_device(display)))
+    if (!(dev = dp_get_device(display)))
 	exit(2);
 
     /* Install a signal handler to restore synaptics parameters on exit */
@@ -624,7 +597,11 @@ main(int argc, char *argv[])
 	    fclose(fd);
 	}
     }
-#ifdef HAVE_XRECORD
+
+    pad_disabled = False;
+    store_current_touchpad_state();
+
+#ifdef HAVE_X11_EXTENSIONS_RECORD_H
     if (use_xrecord)
     {
 	if(check_xrecord(display))
@@ -635,7 +612,7 @@ main(int argc, char *argv[])
             exit(2);
         }
     } else
-#endif /* HAVE_XRECORD */
+#endif /* HAVE_X11_EXTENSIONS_RECORD_H */
       {
 	setup_keyboard_mask(display, ignore_modifier_keys);
 
